@@ -29,7 +29,7 @@ import random
 import sys
 import time
 import traceback
-from mapreduce.third_party import simplejson
+import simplejson
 
 from google.appengine.ext import ndb
 
@@ -43,13 +43,14 @@ from mapreduce import base_handler
 from mapreduce import context
 from mapreduce import errors
 from mapreduce import input_readers
+from mapreduce import map_job_context
 from mapreduce import model
 from mapreduce import operation
 from mapreduce import output_writers
 from mapreduce import parameters
+from mapreduce import shard_life_cycle
 from mapreduce import util
 from mapreduce.api import map_job
-from mapreduce.api.map_job import shard_life_cycle
 from google.appengine.runtime import apiproxy_errors
 
 # pylint: disable=g-import-not-at-top
@@ -129,6 +130,8 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
     """Constructor."""
     super(MapperWorkerCallbackHandler, self).__init__(*args)
     self._time = time.time
+    self.slice_context = None
+    self.shard_context = None
 
   def _drop_gracefully(self):
     """Drop worker task gracefully.
@@ -393,6 +396,19 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       if last_slice:
         obj.end_shard(shard_context)
 
+  def _lc_start_slice(self, tstate, slice_id):
+    self._maintain_LC(tstate.output_writer, slice_id)
+    self._maintain_LC(tstate.input_reader, slice_id)
+    self._maintain_LC(tstate.handler, slice_id)
+
+  def _lc_end_slice(self, tstate, slice_id, last_slice=False):
+    self._maintain_LC(tstate.handler, slice_id, last_slice=last_slice,
+                      begin_slice=False)
+    self._maintain_LC(tstate.input_reader, slice_id, last_slice=last_slice,
+                      begin_slice=False)
+    self._maintain_LC(tstate.output_writer, slice_id, last_slice=last_slice,
+                      begin_slice=False)
+
   def handle(self):
     """Handle request.
 
@@ -449,16 +465,27 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
     job_config = map_job.JobConfig._to_map_job_config(
         spec,
         os.environ.get("HTTP_X_APPENGINE_QUEUENAME"))
-    job_context = map_job.JobContext(job_config)
-    self.shard_context = map_job.ShardContext(job_context, shard_state)
-    self.slice_context = map_job.SliceContext(self.shard_context,
-                                              shard_state,
-                                              tstate)
+    job_context = map_job_context.JobContext(job_config)
+    self.shard_context = map_job_context.ShardContext(job_context, shard_state)
+    self.slice_context = map_job_context.SliceContext(self.shard_context,
+                                                      shard_state,
+                                                      tstate)
     try:
       slice_id = tstate.slice_id
-      self._maintain_LC(tstate.handler, slice_id)
-      self._maintain_LC(tstate.input_reader, slice_id)
-      self._maintain_LC(tstate.output_writer, slice_id)
+      self._lc_start_slice(tstate, slice_id)
+
+      if shard_state.is_input_finished():
+        self._lc_end_slice(tstate, slice_id, last_slice=True)
+        # Finalize the stream and set status if there's no more input.
+        if (tstate.output_writer and
+            isinstance(tstate.output_writer, output_writers.OutputWriter)):
+          # It's possible that finalization is successful but
+          # saving state failed. In this case this shard will retry upon
+          # finalization error.
+          # TODO(user): make finalize method idempotent!
+          tstate.output_writer.finalize(ctx, shard_state)
+        shard_state.set_for_success()
+        return self.__return(shard_state, tstate, task_directive)
 
       if is_this_a_retry:
         task_directive = self._attempt_slice_recovery(shard_state, tstate)
@@ -468,23 +495,15 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       last_slice = self._process_inputs(
           tstate.input_reader, shard_state, tstate, ctx)
 
-      self._maintain_LC(tstate.handler, slice_id, last_slice, False)
-      self._maintain_LC(tstate.input_reader, slice_id, last_slice, False)
-      self._maintain_LC(tstate.output_writer, slice_id, last_slice, False)
+      self._lc_end_slice(tstate, slice_id)
 
       ctx.flush()
 
       if last_slice:
-        # Since there was no exception raised, we can finalize output writer
-        # safely. Otherwise writer might be stuck in some bad state.
-        if (tstate.output_writer and
-            isinstance(tstate.output_writer, output_writers.OutputWriter)):
-          # It's possible that finalization is successful but
-          # saving state failed. In this case this shard will retry upon
-          # finalization error.
-          # TODO(user): make finalize method idempotent!
-          tstate.output_writer.finalize(ctx, shard_state)
-        shard_state.set_for_success()
+        # We're done processing data but we still need to finalize the output
+        # stream. We save this condition in datastore and force a new slice.
+        # That way if finalize fails no input data will be retried.
+        shard_state.set_input_finished()
     # pylint: disable=broad-except
     except Exception, e:
       logging.warning("Shard %s got error.", shard_state.shard_id)
